from typing import Any, Optional, Union, Type, List, Dict

import torch


class ModelDeploymentConfig:
    __slots__ = ('model_class', 'path', 'port', 'allocations', 'model_kwargs', 'use_torchscript', 'max_batch_size')

    def __init__(
            self, model_class: Type[torch.nn.Module], path: str, port: int,
            allocations: Union[List[int], str], model_kwargs: Optional[Dict[str, Any]] = None,
            use_torchscript: bool = True, max_batch_size: Optional[int] = None
    ):
        self.model_class = model_class
        self.path = path
        self.model_kwargs: Dict[str, Any] = {} if model_kwargs is None else model_kwargs
        self.port = port
        self.allocations = list(map(int, allocations.split(','))) if type(allocations) == str else allocations
        self.use_torchscript = use_torchscript
        self.max_batch_size = max_batch_size



import sys
import logging


def init_logger():
    service_logger = logging.getLogger('fastformers')
    service_logger.propagate = False
    service_logger.setLevel(logging.DEBUG)
    if service_logger.handlers:
        return service_logger

    # create and set formatting for the logging file handler
    fh = logging.StreamHandler(sys.stdout)
    fh.setFormatter(logging.Formatter(fmt="%(asctime)s\t%(message)s", datefmt="%H:%M:%S"))

    # add handler to logger object
    service_logger.addHandler(fh)
    return service_logger


logger = init_logger()



from typing import Any, Optional, Union, Tuple, List, Dict

import time
import select
import socket
import threading
import operator
import uuid
import pickle as pkl
import torch

from ..utils import logger
from .socket_connection import SocketConnection
from .model_deployment_config import ModelDeploymentConfig
from .configs import Config


class ClientSocket(SocketConnection):
    __slots__ = ('message_id', 'batch_ind')

    def __init__(self, connection: socket.socket, message_size: int):
        super().__init__(connection, message_size)
        self.connection.setblocking(False)
        self.message_id: Optional[bytes] = None
        self.batch_ind: int = -1

    def empty(self):
        self.connection.recv(self.message_size)

    def request_server(self):
        self.connection.send(Config.request_signal)

    def cancel_request(self):
        self.connection.send(Config.request_cancelled)

    def send_inputs(self, data: Any, batch_ind: int):
        self.empty()
        self.message_id = uuid.uuid4().bytes[:Config.message_id_len]
        self.send_message(self.message_id, data)
        self.batch_ind = batch_ind

    def receive_outputs(self) -> Optional[Tuple[int, Any]]:
        self.connection.setblocking(True)
        data = self.read_buffer()
        self.connection.setblocking(False)
        response_message_id = data[:Config.message_id_len]
        if response_message_id != self.message_id:
            logger.warning(f'Skipping response with wrong message id')
            return None
        data = data[Config.message_id_len:]
        self.message_id = None
        if data.startswith(Config.error_prefix):
            raise ConnectionError(data[len(Config.error_prefix):].decode('utf-8'))
        return self.batch_ind, pkl.loads(data)


class ClientAPI:
    client_sockets: Dict[bytes, Dict[str, List[ClientSocket]]] = {}

    def __init__(self, model: str, model_config: ModelDeploymentConfig):
        self.model = model
        self.model_config = model_config

    @staticmethod
    def create_sockets(
            model: str, model_config: ModelDeploymentConfig, attempts: int, wait_time: float) -> List[ClientSocket]:
        logger.debug('Connecting to model workers')
        model_sockets: List[ClientSocket] = []
        for worker_ind in range(len(model_config.allocations)):
            client_socket = socket.socket()
            for _ in range(attempts):
                try:
                    client_socket.connect((Config.host, model_config.port + worker_ind))
                    logger.debug(f'{model} model worker {worker_ind} connected')
                    model_sockets.append(
                        ClientSocket(connection=client_socket, message_size=Config.client_message_size))
                    break
                except ConnectionRefusedError:
                    time.sleep(wait_time)
        return model_sockets

    @staticmethod
    def split_inputs_to_batches(inputs: Dict, max_batch_size: int) -> List[Dict]:
        batch_size = next(iter(inputs.values())).shape[0]
        if max_batch_size is None or batch_size <= max_batch_size:
            return [inputs]
        return [
            {key: value[ind: ind + max_batch_size] for key, value in inputs.items()}
            for ind in range(0, batch_size, max_batch_size)
        ]

    @staticmethod
    def concat_lists_or_tensors(items: Union[List[List], List[torch.Tensor]]) -> Union[List, torch.Tensor]:
        return (
            torch.cat(items) if isinstance(items[0], torch.Tensor)
            else [value for sample_values in items for value in sample_values]
        )

    @classmethod
    def stack_outputs(cls, outputs: List) -> Any:
        sample = outputs[0]
        if len(outputs) == 1:
            return sample
        if isinstance(sample, dict):
            return {
                key: cls.concat_lists_or_tensors(list(map(operator.itemgetter(key), outputs))) for key in sample.keys()
            }
        return cls.concat_lists_or_tensors(outputs)

    def forward(self, **inputs):
        logger.info('Inside client forward')
        thread_id = threading.get_ident().to_bytes(length=Config.thread_id_len, byteorder='big', signed=False)

        if thread_id not in self.client_sockets:
            self.client_sockets[thread_id] = {}
        if self.model not in self.client_sockets[thread_id]:
            self.client_sockets[thread_id][self.model] = self.create_sockets(
                self.model, self.model_config, attempts=120, wait_time=10)

        client_sockets = self.client_sockets[thread_id][self.model]

        logger.info('Sockets acquired')

        buffered_sockets = select.select(client_sockets, [], [], 0)[0]

        for client_socket in buffered_sockets:
            client_socket.empty()

        logger.info('Sockets cleared')

        batched_inputs = self.split_inputs_to_batches(inputs, self.model_config.max_batch_size)

        outputs: List = [None] * len(batched_inputs)
        received_counter = 0
        cur_ind = 0

        logger.info(f'Num batches: {len(batched_inputs)}')

        for client_socket in client_sockets:
            client_socket.request_server()

        while received_counter < len(outputs):
            logger.info('Waiting for select')

            ready_sockets: List[ClientSocket] = select.select(
                client_sockets if batched_inputs else
                [client_socket for client_socket in client_sockets if client_socket.message_id is not None],
                [], [], None
            )[0]
            input_sockets: List[ClientSocket] = []
            output_sockets: List[ClientSocket] = []
            for client_socket in ready_sockets:
                (input_sockets if client_socket.message_id is None else output_sockets).append(client_socket)

            added_connections = min(len(input_sockets), len(batched_inputs))

            selected_sockets: List[ClientSocket] = input_sockets[:added_connections]

            for ind, (inputs, client_socket) in enumerate(zip(batched_inputs, selected_sockets)):
                logger.info('Sending inputs')
                client_socket.send_inputs(inputs, cur_ind + ind)

            cur_ind += added_connections

            batched_inputs = batched_inputs[added_connections:]

            if added_connections and not batched_inputs:
                logger.info('Cancelling requests')
                for client_socket in client_sockets:
                    if client_socket.message_id is None:
                        client_socket.cancel_request()

            for client_socket in output_sockets:
                logger.info('Waiting to receive')
                socket_outputs = client_socket.receive_outputs()
                logger.info('Received replies')
                if socket_outputs is None:
                    logger.info('Skipping')
                    continue
                batch_ind, data = socket_outputs
                outputs[batch_ind] = data
                received_counter += 1
                if batched_inputs:
                    logger.info('Requesting server again')
                    client_socket.request_server()

        logger.info('Stacking outputs')

        return self.stack_outputs(outputs)

    def __call__(self, **kwargs):
        return self.forward(**kwargs)

    def close(self):
        for client_sockets in self.client_sockets.values():
            for sockets in client_sockets.values():
                for client_socket in sockets:
                    client_socket.close()



class Config:
    host = '127.0.0.1'

    end_delimiter = b'DELIM'

    socket_timeout = 20

    server_message_size = 8192

    client_message_size = 65536

    error_prefix = b'ERROR'

    thread_id_len = 8
    message_id_len = 4

    request_signal = b'\x00'
    request_cancelled = b'\x01'



from typing import Any

import socket
import pickle as pkl

from ..utils import logger
from .configs import Config


class SocketConnection:
    __slots__ = ('connection', 'message_size', 'alive')

    delimiter = Config.end_delimiter

    def __init__(self, connection: socket.socket, message_size: int):
        self.connection = connection
        self.connection.settimeout(Config.socket_timeout)
        self.message_size = message_size
        self.alive = True

    def fileno(self) -> int:
        return self.connection.fileno()

    def close(self):
        if self.alive:
            logger.debug(f'Closing connection {self.fileno()}')
            self.alive = False
            self.connection.close()

    def read_buffer(self) -> bytes:
        response = bytes()
        while True:
            buffer = self.connection.recv(self.message_size)
            if not buffer:
                raise ConnectionError('No response, other side probably died')
            response += buffer
            if response.endswith(self.delimiter):
                return response.split(self.delimiter)[-2]

    def receive_message(self) -> Any:
        if not self.alive:
            raise ConnectionError('Trying to receive message over closed connection')
        try:
            data = self.read_buffer()
            if data.startswith(Config.error_prefix):
                raise ConnectionError(data[len(Config.error_prefix):].decode('utf-8'))
            return pkl.loads(data)
        except (pkl.UnpicklingError, ConnectionError):
            self.close()
            raise

    def send_data(self, data: bytes):
        if not self.alive:
            raise ConnectionError('Trying to send message over closed connection')
        try:
            self.connection.send(data + self.delimiter)
        except ConnectionError:
            self.close()
            raise

    def send_message(self, message_id: bytes, data: Any):
        self.send_data(message_id + pkl.dumps(data))

    def send_error(self, message_id: bytes, error: str):
        self.send_data(message_id + Config.error_prefix + error.encode('utf-8'))



from ._version import __version__

from .models import (
    BartForSequenceClassification, DistilBertForMaskedLM, ElectraForSequenceClassification, RobertaForQuestionAnswering,
    CoreferencePredictor
)

from .multiprocess_api import ClientAPI, ModelDeploymentConfig

from .run_model_fleet import run_model_fleet



import torch


class TimeDistributed(torch.nn.Module):
    def __init__(self, module):
        super().__init__()
        self._module = module

    def get_input_dim(self) -> int:
        return self._module.get_input_dim()

    def get_output_dim(self) -> int:
        return self._module.get_output_dim()

    def forward(self, inputs):
        outputs = self._module(inputs.flatten(0, 1))
        return outputs.view(inputs.shape[:2] + outputs.shape[1:])



from typing import Optional, Union, List, Dict, Iterable

import select
import socket
import time
import pickle as pkl

import torch.jit

from .configs import Config
from ..utils import logger

from .socket_connection import SocketConnection


class ServerSocket(SocketConnection):
    def read_buffer(self) -> Optional[bytes]:
        response = bytes()
        while True:
            buffer = self.connection.recv(self.message_size)
            if not buffer:
                raise ConnectionError('No response, other side probably died')
            response += buffer
            if response.startswith(Config.request_cancelled):
                return None
            if response.endswith(self.delimiter):
                return response.split(self.delimiter)[-2]

    def receive_message(self):
        try:
            response = self.connection.recv(self.message_size)
            if not response:
                raise ConnectionError('No response, other side probably died')
        except ConnectionError:
            self.alive = False
            raise
        active_request = response.endswith(Config.request_signal)
        if active_request:
            self.connection.send(Config.request_signal)
            response = self.read_buffer()
            if response is None:
                return None
            return response[:Config.message_id_len], pkl.loads(response[Config.message_id_len:])


class SessionState:
    def __init__(self, connection: SocketConnection, message_id: bytes, inputs: dict):
        self.connection = connection
        self.message_id = message_id
        self.inputs = inputs

    def send_output(self, output):
        try:
            self.connection.send_message(self.message_id, output)
        except ConnectionError:
            pass

    def send_error(self, e: Exception):
        try:
            self.connection.send_error(self.message_id, str(e))
        except ConnectionError:
            pass


class ModelServer:
    def __init__(self, port: int, model: torch.nn.Module, model_name: str):
        logger.info('Server starting')
        self.model = model
        self.device = next(self.model.parameters()).device
        self.server_socket = self.create_server_socket(port)
        logger.info(f'{model_name} model server started on port {port}')

    @staticmethod
    def create_server_socket(port: int) -> socket.socket:
        server_socket = socket.socket()
        while True:
            try:
                server_socket.bind((Config.host, port))
                break
            except OSError:
                time.sleep(5)
        server_socket.listen(10)
        return server_socket

    def close(self):
        self.server_socket.close()

    @staticmethod
    def pad_and_concat(tensors: List[torch.Tensor], pad: int = 0) -> torch.Tensor:
        sample_tensor = tensors[0]
        if len(tensors) == 1:
            return sample_tensor

        max_len = max([tensor.shape[1] for tensor in tensors])

        total_batch_size = 0
        for tensor in tensors:
            total_batch_size += tensor.shape[0]
        output_tensor = torch.empty((total_batch_size, max_len) + sample_tensor.shape[2:],
                                    dtype=sample_tensor.dtype, device=sample_tensor.device).fill_(pad)
        cur_ind = 0
        for tensor in tensors:
            new_ind = cur_ind + tensor.shape[0]
            output_tensor[cur_ind: new_ind, :tensor.shape[1]] = tensor
            cur_ind = new_ind

        return output_tensor

    @staticmethod
    def split_tensor_output(outputs: torch.Tensor, sizes: List[int]) -> Iterable[torch.Tensor]:
        outputs = outputs.detach().to('cpu')
        last_ind = 0
        for size in sizes:
            new_ind = last_ind + size
            yield outputs[last_ind: new_ind]
            last_ind = new_ind

    @staticmethod
    def split_dict_output(
            outputs: Dict[str, Union[list, torch.Tensor]], sizes: List[int]
    ) -> Iterable[Dict[str, Union[list, torch.Tensor]]]:
        outputs = {
            key: value.detach().to('cpu') if isinstance(value, torch.Tensor) else value
            for key, value in outputs.items()
        }
        last_ind = 0
        for size in sizes:
            new_ind = last_ind + size
            yield {key: value[last_ind: new_ind] for key, value in outputs.items()}
            last_ind = new_ind

    def split_output(self, outputs, sizes: List[int]) -> Iterable:
        return (self.split_tensor_output(outputs, sizes) if isinstance(outputs, torch.Tensor)
                else self.split_dict_output(outputs, sizes))

    def run(self):
        connections: List[Union[SocketConnection, socket.socket]] = []

        while True:
            connections = [connection for connection in connections if connection.alive]

            logger.info('Waiting for connections')
            new_connections = select.select(connections + [self.server_socket], [], [], None)[0]
            logger.info('Connection received')

            if self.server_socket in new_connections:
                connection, address = self.server_socket.accept()
                connection = ServerSocket(connection, Config.server_message_size)
                logger.debug(f'Added connection, fileno: {connection.fileno()}')
                connections.append(connection)
                continue

            session_states: List[SessionState] = []
            for connection in new_connections:
                try:
                    inputs = connection.receive_message()
                except ConnectionError:
                    logger.debug('ConnectionError')
                    continue
                if not inputs:
                    logger.debug('Empty')
                    continue
                message_id, input_dict = inputs
                session_states.append(SessionState(connection, message_id, input_dict))
                logger.debug('Processed new inputs')

            if not session_states:
                continue

            inputs = [session_state.inputs for session_state in session_states]
            pad_map: Dict[str, int] = getattr(self.model, 'pad_map', {})

            try:
                keys = list(inputs[0].keys())
                sizes = [sample_inputs[keys[0]].shape[0] for sample_inputs in inputs]
                inputs = {
                    key: self.pad_and_concat(
                        [sample_inputs[key].to(self.device) for sample_inputs in inputs], pad_map.get(key, 0))
                    for key in keys
                }
                with torch.no_grad():
                    outputs = self.model.forward(**inputs)
                logger.debug('Sending outputs')
                for session_state, output in zip(session_states, self.split_output(outputs, sizes)):
                    session_state.send_output(output)
            except Exception as e:
                logger.exception(e)
                for session_state in session_states:
                    session_state.send_error(e)



import setuptools
import os
import re

package_name = "fastformers"
package_dir = "src"


def parse_version() -> str:
    with open(os.path.join(package_dir, package_name, "_version.py"), "rt") as fh:
        version_contents = fh.read()
    match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_contents, re.M)
    if match:
        return match.group(1)
    else:
        raise RuntimeError("Unable to find version string")


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name=package_name,
    version=parse_version(),
    author="datarootlabs",
    author_email="dev@datarootlabs.com",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    package_dir={"": package_dir},
    packages=setuptools.find_packages(where=package_dir),
    python_requires=">=3.6",
    install_requires=[
        'torch>=1.6',
    ]
)



from typing import Optional

import math

import torch
from torch.nn import Module, ModuleList, Linear, Embedding, LayerNorm
from torch.nn.functional import gelu


from ..utils import neg_inf


class MultiHeadAttention(Module):
    def __init__(self, n_heads: int, dim: int):
        super().__init__()
        self.n_heads = n_heads
        self.dim_per_head = dim // n_heads
        self.scale = math.sqrt(self.dim_per_head)

        self.pre_attention = Linear(dim, dim * 3)
        self.out_lin = Linear(dim, dim)

    def forward(self, query: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        batch_size, seq_len, dim = query.shape

        query = self.pre_attention(query)
        query = query.view(batch_size, seq_len, self.n_heads, self.dim_per_head, 3)
        query = query.transpose(1, 2).contiguous()
        query = query.view(batch_size * self.n_heads, seq_len, self.dim_per_head, 3)

        q = query[:, :, :, 0] / self.scale
        k_v = query[:, :, :, 1:]

        dot_prod = q.bmm(k_v[:, :, :, 0].transpose(1, 2))

        dot_prod.masked_fill_(mask, neg_inf(dot_prod.dtype == torch.float16))

        attn_weights = dot_prod.softmax(dim=-1, dtype=torch.float).type_as(query)

        out = self.out_lin(attn_weights.bmm(k_v[:, :, :, 1]).view(
            batch_size, self.n_heads, seq_len,
            self.dim_per_head).transpose(1, 2).contiguous().view(batch_size, seq_len, dim))

        return out


class TransformerFFN(Module):
    def __init__(self, dim: int, dim_hidden: int):
        super(TransformerFFN, self).__init__()
        self.lin1 = Linear(dim, dim_hidden)
        self.lin2 = Linear(dim_hidden, dim)

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        tensor = self.lin1(tensor)
        tensor = gelu(tensor)
        tensor = self.lin2(tensor)
        return tensor


class TransformerEncoderLayer(Module):
    def __init__(self, n_heads: int, embedding_size: int, ffn_size: int):
        super().__init__()
        self.attention = MultiHeadAttention(n_heads, embedding_size)
        self.norm1 = LayerNorm(embedding_size)
        self.ffn = TransformerFFN(embedding_size, ffn_size)
        self.norm2 = LayerNorm(embedding_size)

    def forward(self, tensor: torch.Tensor, mask: torch.Tensor):
        residual = tensor
        tensor = self.attention(tensor, mask=mask)
        tensor = tensor + residual
        tensor = self.norm1(tensor)
        tensor = tensor + self.ffn(tensor)
        tensor = self.norm2(tensor)
        return tensor


class TransformerEncoder(Module):
    def __init__(self, n_heads: int, n_layers: int, embedding_size: int, ffn_size: int):
        super(TransformerEncoder, self).__init__()

        self.layers = ModuleList(TransformerEncoderLayer(n_heads, embedding_size, ffn_size) for _ in range(n_layers))
        self.n_heads = n_heads

    def forward(self, tensor: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        bsz, seq_len = mask.shape
        attn_mask = ~mask.view(bsz, 1, -1, seq_len).repeat(1, self.n_heads, 1, 1).expand(
            -1, -1, seq_len, -1).view(bsz * self.n_heads, seq_len, seq_len)

        for layer in self.layers:
            tensor = layer.forward(tensor, attn_mask)

        return tensor * mask.unsqueeze(-1)


class TransformerDecoder(Module):
    def __init__(self, n_heads: int, n_layers: int, embedding_size: int, ffn_size: int):
        super().__init__()

        self.layers = ModuleList(TransformerDecoderLayer(n_heads, embedding_size, ffn_size) for _ in range(n_layers))
        self.n_heads = n_heads

    def process_decoder_mask(self, decoder_mask: torch.Tensor, decoder_length: int) -> torch.Tensor:
        bsz, seq_len = decoder_mask.shape
        mask = torch.tril(torch.ones(seq_len, seq_len, dtype=torch.bool, device=decoder_mask.device))
        mask = mask.unsqueeze(0)[:, -decoder_length:] & decoder_mask.unsqueeze(1)
        return ~mask.view(bsz, 1, -1, seq_len).repeat(1, self.n_heads, 1, 1).expand(
            -1, -1, decoder_length, -1).view(bsz * self.n_heads, decoder_length, seq_len)

    def process_encoder_mask(self, encoder_mask: torch.Tensor, decoder_length: int) -> torch.Tensor:
        bsz, encoder_length = encoder_mask.shape
        return ~encoder_mask.view(bsz, 1, -1, encoder_length).repeat(1, self.n_heads, 1, 1).expand(
            -1, -1, decoder_length, -1).view(bsz * self.n_heads, decoder_length, encoder_length)

    def forward(
            self, tensor: torch.Tensor, decoder_mask: torch.Tensor,
            encoder_mask: torch.Tensor, encoder_state: torch.Tensor
    ) -> torch.Tensor:

        decoder_length = tensor.shape[1]
        decoder_mask = self.process_decoder_mask(decoder_mask, decoder_length=decoder_length)
        encoder_mask = self.process_encoder_mask(encoder_mask, decoder_length=decoder_length)

        encoder_state = encoder_state.flatten(0, 1)

        for ind, layer in enumerate(self.layers):
            tensor = layer.forward(
                x=tensor, decoder_mask=decoder_mask, encoder_state=encoder_state[:, ind], encoder_mask=encoder_mask
            )

        return tensor


class TransformerDecoderLayer(Module):
    def __init__(self, n_heads: int, embedding_size: int, ffn_size: int):
        super().__init__()
        self.dim = embedding_size

        self.self_attention = MultiHeadAttention(n_heads, embedding_size)
        self.norm1 = LayerNorm(embedding_size)

        self.encoder_attention = DecoderEncoderAttention(n_heads, embedding_size)
        self.norm2 = LayerNorm(embedding_size)

        self.ffn = TransformerFFN(embedding_size, ffn_size)
        self.norm3 = LayerNorm(embedding_size)

    def forward(
            self, x, decoder_mask: torch.Tensor, encoder_state: torch.Tensor, encoder_mask: torch.Tensor
    ) -> torch.Tensor:
        residual = x

        x = self.self_attention(query=x, mask=decoder_mask)

        x = x + residual
        x = self.norm1(x)

        x = self.encoder_attention(query=x, key=encoder_state[:, 0], value=encoder_state[:, 1], mask=encoder_mask) + x

        x = self.norm2(x)

        residual = x

        x = self.ffn(x)

        x = residual + x
        x = self.norm3(x)

        return x


class DecoderEncoderAttention(Module):
    def __init__(self, n_heads: int, dim: int):
        super().__init__()
        self.n_heads = n_heads
        self.dim_per_head = dim // n_heads
        self.scale = math.sqrt(self.dim_per_head)
        self.q_lin = Linear(dim, dim)
        self.out_lin = Linear(dim, dim)

    def forward(self, query: torch.Tensor, mask: torch.Tensor, key: torch.Tensor, value: torch.Tensor) -> torch.Tensor:
        batch_size, seq_len, dim = query.shape

        query = self.q_lin(query) / self.scale
        query = query.view(batch_size, seq_len, self.n_heads, self.dim_per_head).transpose(1, 2).contiguous()
        query = query.view(batch_size * self.n_heads, seq_len, self.dim_per_head)

        dot_prod = query.bmm(key.transpose(1, 2))

        dot_prod.masked_fill_(mask, neg_inf(dot_prod.dtype == torch.float16))

        attn_weights = dot_prod.softmax(dim=-1, dtype=torch.float).type_as(query)

        return self.out_lin(attn_weights.bmm(value).view(
            batch_size, self.n_heads, seq_len,
            self.dim_per_head).transpose(1, 2).contiguous().view(batch_size, seq_len, dim))


class Embeddings(Module):
    def __init__(self, vocab_size: int, dim: int, padding_idx: int, max_position_embeddings: int):
        super().__init__()
        self.word_embeddings = Embedding(vocab_size, dim, padding_idx=padding_idx)
        self.position_embeddings = Embedding(max_position_embeddings, dim)
        self.LayerNorm = LayerNorm(dim)

    def forward(self, input_ids: torch.Tensor) -> torch.Tensor:
        embeddings = self.word_embeddings(input_ids) + self.position_embeddings.weight[:input_ids.shape[1]]
        return self.LayerNorm(embeddings)


class MaskEmbeddings(Embeddings):
    def __init__(self, vocab_size: int, dim: int, padding_idx: int, max_position_embeddings: int):
        super().__init__(vocab_size, dim, padding_idx, max_position_embeddings)
        self.padding_idx = padding_idx

    @staticmethod
    def create_position_ids_from_input_ids(input_ids: torch.Tensor, padding_idx: int) -> torch.Tensor:
        mask = input_ids.ne(padding_idx)
        return torch.cumsum(mask, dim=1) * mask + padding_idx

    def forward(self, input_ids: torch.Tensor) -> torch.Tensor:
        position_ids = self.create_position_ids_from_input_ids(input_ids, self.padding_idx)
        embedded = self.word_embeddings(input_ids) + self.position_embeddings(position_ids)
        return self.LayerNorm(embedded)


class EmbeddingsWithTokenTypes(Module):
    def __init__(
            self, vocab_size: int, token_types: int, embedding_size: int, pad_token_id: int,
            max_position_embeddings: int
    ):
        super().__init__()
        self.word_embeddings = Embedding(vocab_size, embedding_size, padding_idx=pad_token_id)
        self.position_embeddings = Embedding(max_position_embeddings, embedding_size)
        self.token_type_embeddings = Embedding(token_types, embedding_size)
        self.LayerNorm = LayerNorm(embedding_size)

    def forward(self, input_ids: torch.Tensor, token_type_ids: Optional[torch.Tensor]):
        embeddings = self.word_embeddings(input_ids) + self.position_embeddings.weight[:input_ids.shape[1]] + (
            self.token_type_embeddings.weight[0] if token_type_ids is None
            else self.token_type_embeddings(token_type_ids))
        return self.LayerNorm(embeddings)


class TanhHead(Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, input_dim: int, inner_dim: int, num_classes: int):
        super().__init__()
        self.dense = Linear(input_dim, inner_dim)
        self.out_proj = Linear(inner_dim, num_classes)

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = torch.tanh(hidden_states)
        return self.out_proj(hidden_states)
